from libc.stdint cimport uint32_t
from libc.stdint cimport uint64_t

#from thinc.features cimport feat_t
from cymem.cymem cimport Pool
from .maps cimport key_t
from .maps cimport MapStruct


ctypedef uint32_t idx_t
ctypedef uint32_t feat_t


cdef struct Node:
    Node* nodes
    idx_t offset
    idx_t length
    idx_t value


cdef class SequenceIndex:
    cdef Pool mem
    cdef Node* tree
    cdef MapStruct* pmap
    cdef readonly idx_t i
    cdef readonly idx_t longest_node 
    cdef idx_t get(self, feat_t* feat, size_t n) except *
    cdef idx_t index(self, feat_t* feat, size_t n) except 0



import ujson

from libc.string cimport memcpy

from murmurhash.mrmr cimport hash64
from cymem.cymem cimport Address

from .maps cimport map_init, map_get, map_set


DEF MAX_TRIE_VALUE = 100000


cdef class SequenceIndex:
    """Map numeric arrays to sequential unsigned integers, starting from 1. 0
    indicates that the array was not found.
    """
    def __init__(self, idx_t offset=0):
        self.mem = Pool()
        self.tree = <Node*>self.mem.alloc(1, sizeof(Node))
        self.pmap = <MapStruct*>self.mem.alloc(1, sizeof(MapStruct))
        map_init(self.mem, self.pmap, 8)
        # Value of 0 set aside for special use by the parent code, whatever that
        # might be.
        self.i = 1 + offset
        self.longest_node = 0
        assert self.tree.nodes is NULL

    def __getitem__(self, feature):
        try:
            len(feature)
        except TypeError:
            feature = (feature,)
        cdef Address mem = array_from_seq(feature)
        return self.get(<feat_t*>mem.ptr, len(feature))

    def __call__(self, *feature):
        cdef Address mem = array_from_seq(feature)
        return self.index(<feat_t*>mem.ptr, len(feature))

    cdef idx_t get(self, feat_t* feature, size_t n) except *:
        # First, check we don't have any over-size values
        cdef idx_t i
        cdef key_t hashed
        cdef Node* node = self.tree
        cdef feat_t f
        for i in range(n):
            f = feature[i]
            if f >= MAX_TRIE_VALUE:
                hashed = hash64(feature, n * sizeof(feat_t), 0)
                return <idx_t>map_get(self.pmap, hashed)
            elif node.offset <= f < (node.offset + node.length):
                node = &node.nodes[f - node.offset]
            else:
                return 0
        return node.value

    cdef idx_t index(self, feat_t* feature, size_t n) except 0:
        assert n >= 1
        cdef key_t hashed
        cdef idx_t i
        for i in range(n):
            if feature[i] >= MAX_TRIE_VALUE:
                hashed = hash64(feature, n * sizeof(feat_t), 0)
                map_set(self.mem, self.pmap, hashed, <void*>self.i)
                self.i += 1
                return self.i - 1

        cdef Node* node = self.tree
        cdef feat_t f
        for i in range(n):
            f = feature[i]
            if node.nodes == NULL:
                node.offset = f
                node.length = 1
                node.nodes = <Node*>self.mem.alloc(node.length, sizeof(Node))
            elif f < node.offset:
                node.nodes = array_prepend(self.mem, node.nodes, node.length,
                                           node.offset - f)
                node.length += node.offset - f
                node.offset = f
            elif f >= (node.length + node.offset):
                node.length = f - node.offset + 1
                node.nodes = <Node*>self.mem.realloc(node.nodes, node.length * sizeof(Node))
            if node.length > self.longest_node:
                self.longest_node = node.length
            node = &(node.nodes[f - node.offset])
        if node.value == 0:
            node.value = self.i
            self.i += 1
        return node.value

    def revalue(self, list new_values):
        assert len(new_values) == self.i
        cdef Address table_mem = array_from_seq(new_values)
        cdef feat_t* table = <feat_t*>table_mem.ptr
        cdef NodeIterator nodes = NodeIterator(self)
        cdef size_t addr
        cdef Node* node
        for _ in nodes:
            nodes.node.value = table[nodes.node.value]
        for i in range(self.pmap.length):
            if self.pmap.cells[i].key != 0:
                self.pmap.cells[i].value = <void*>table[<feat_t>self.pmap.cells[i].value]

    def dump(self, file_):
        for i in range(self.pmap.length):
            if self.pmap.cells[i].key != 0:
                file_.write(_fmt_line(<idx_t>self.pmap.cells[i].value,
                                      [self.pmap.cells[i].key]))
        for address, value in NodeIterator(self):
            if value == 0:
                continue
            file_.write(_fmt_line(value, address))

    def load(self, file_):
        for line in file_:
            value, address = _parse_line(line)
            if not address:
                continue
            self.i = value
            self(*address)


def _fmt_line(value, address):
    return ujson.dumps((value, (address,)))

def _parse_line(line):
    return ujson.loads(line)


cdef class NodeIterator:
    cdef Pool mem
    cdef Node* node
    cdef Node** stack
    cdef size_t* offsets
    cdef idx_t i
    def __init__(self, SequenceIndex trie):
        self.mem = Pool()
        self.stack = <Node**>self.mem.alloc(trie.i, sizeof(Node*))
        self.offsets = <size_t*>self.mem.alloc(trie.i, sizeof(size_t))
        self.stack[0] = trie.tree
        self.i = 1

    def __iter__(self):
        """Iterate over the keys and values in the trie.
        """
        cdef Node* node
        cdef Node* parent
        cdef size_t i
        while self.i != 0:
            self.i -= 1
            node = self.stack[self.i]
            key = []
            i = self.i
            # In parallel to the stack, we note the offset of the node's parent in
            # the stack, so that we can always find the parent of stack-node i at
            # self.stack[i - self.offsets[i]].
            while i != 0:
                parent = self.stack[i - self.offsets[i]]
                key.insert(0, parent.offset + self.offsets[i])
                if self.offsets[i] == 0:
                    i -= 1
                else:
                    i -= self.offsets[i]
            # Push the children onto the stack
            for i in range(node.length):
                self.stack[self.i] = &node.nodes[i]
                self.offsets[self.i] = i    
                self.i += 1
            self.node = node
            yield key, node.value


cdef Address array_from_seq(object seq):
    cdef Address mem = Address(len(seq), sizeof(feat_t))
    array = <feat_t*>mem.ptr
    cdef int i
    cdef feat_t f
    for i, f in enumerate(seq):
        array[i] = f
    return mem


cdef Node* array_prepend(Pool mem, Node* old, size_t length, size_t to_add) except NULL:
    cdef size_t i, j
    cdef Node* new = <Node*>mem.alloc(length + to_add, sizeof(Node))
    memcpy(&new[to_add], old, length * sizeof(Node))
    mem.free(old)
    return new



"""Count occurrences of uint64-valued keys."""

cdef class PreshCounter:
    def __init__(self, initial_size=8):
        assert initial_size != 0
        assert initial_size & (initial_size - 1) == 0
        self.mem = Pool()
        self.c_map = <MapStruct*>self.mem.alloc(1, sizeof(MapStruct))
        map_init(self.mem, self.c_map, initial_size)

    property length:
        def __get__(self):
            return self.c_map.length

    def __len__(self):
        return self.c_map.length

    def __iter__(self):
        cdef int i
        for i in range(self.c_map.length):
            if self.c_map.cells[i].key != 0:
                yield (self.c_map.cells[i].key, <count_t>self.c_map.cells[i].value)

    def __getitem__(self, key_t key):
        return <count_t>map_get(self.c_map, key)

    cpdef int inc(self, key_t key, count_t inc) except -1:
        cdef count_t c = <count_t>map_get(self.c_map, key)
        c += inc
        map_set(self.mem, self.c_map, key, <void*>c)





#!/usr/bin/env python
from distutils.core import setup
from Cython.Build import cythonize
from Cython.Distutils import Extension
from glob import glob


import sys
import os
from os import path
from os.path import splitext

from distutils.sysconfig import get_python_inc

virtual_env = os.environ.get('VIRTUAL_ENV', '')

includes = []

if 'VIRTUAL_ENV' in os.environ:
    includes += glob(path.join(os.environ['VIRTUAL_ENV'], 'include', 'site', '*'))
else:
    # If you're not using virtualenv, set your include dir here.
    pass


exts = [Extension("preshed.maps", ["preshed/maps.pyx"], include_dirs=includes),
        Extension("preshed.tries", ["preshed/tries.pyx"], include_dirs=includes),
        Extension("preshed.counter", ["preshed/counter.pyx"], include_dirs=includes)]

setup(
    ext_modules=cythonize(exts),
    name="preshed",
    packages=["preshed"],
    version="0.20",
    author="Matthew Honnibal",
    author_email="honnibal@gmail.com",
    url="http://github.com/syllog1sm/preshed",
    package_data={"preshed": ["*.pxd", "*.pyx", "*.c"]},
    description="""Cython hash table that trusts the keys are pre-hashed""",
    classifiers=[
                'Environment :: Console',
                'Operating System :: OS Independent',
                'Intended Audience :: Science/Research',
                'Programming Language :: Cython',
                'Topic :: Scientific/Engineering'],
)
